#'@title Execution mode for building and/or classification
#'@description Sets the execution mode for building and/or classification process, 0 = GPU, 1 = sequential, 2 = multithreaded.
#'@param build Value for the building process.
#'@param class Value for the training process.
#'@examples
#' eccCLsetExecutionMode(2,0) : Sets building to multi-threaded CPU execution
#' and classification to GPU execution.
eccCLsetExecutionMode <- function(build, class) {
  .jcall("ecccl.RBridge", "V", "setBuildExecution", as.integer(build))
  .jcall("ecccl.RBridge", "V", "setClassExecution", as.integer(class))
}

#'@title OpenCL library name
#'@description In case the native OpenCL library isn't named libOpenCL.so, the package won't find it.
#'With this method, a different name can be set to look for.
#'@param libName String for the library name to look for.
#'@examples
#'eccCLsetCLLibname("libOpenCL.so.1")
eccCLsetCLLibname <- function(libName) {
  oldLib <- .jcall("java/lang/System","S","setProperty","org.lwjgl.opencl.libname", libName)
  cat("changed org.lwjgl.opencl.libname to", libName, " (old value: ", oldLib, ")")
}

#'@title Loads a weka-formatted file.
#'@description A weka-file consists of a pair of files. One has the ending .arff and the other one .xml. This method loads
#'these files and returns an internal representation of it as a Java-object.
#'@param wekaFile File as string and path to the file (excluding .arff and .xml)
#'@return Representation of the file as Java-object.
#'@examples
#'weka <- eccCLloadWekaFile("/home/user/FILE") : Loads /home/user/FILE.arff and /home/user/FILE.xml
eccCLloadWekaFile <- function(wekaFile) {
  loader <- .jnew("ecccl.WekaFrameLoader")
  .jcall(loader, "V", "load", wekaFile)

  return (loader)
}

#'@title Extracts the attributes out of a wekafile.
#'@description Extracts the attributes out of a wekafile and returns them as a data.frame.
#'The data.frame consists of (number of instances) rows and (number of attributes) columns.
#'Each row represents the attributes an instance.
#'@param wekaLoader A wekafile loaded e.g. with \code{\link{eccCLloadWekaFile}}.
#'@return A data.frame containing all attributes of a wekafile.
#'@examples
#'values <- eccCLwekaValues(weka)
eccCLwekaValues <- function(wekaLoader) {
  numInstances <- .jcall(wekaLoader, "I", "getNumInstances")
  numAttributes <- .jcall(wekaLoader, "I", "getNumAttributes")
  values <- .jcall(wekaLoader, "[D", "getAttributes")
  dim(values) <- c(numInstances, numAttributes)
  valueFrame = as.data.frame(values)
  return (valueFrame)
}

#'@title Extracts the labels out of a wekafile.
#'@description Extracts the labels out of a wekafile and returns them as a data.frame.
#'The data.frame consists of (number of instances) rows and (number of labels) columns.
#'Each row represents the labels of an instance.
#'@param wekaLoader A wekafile loaded e.g. with \code{\link{eccCLloadWekaFile}}.
#'@return A data.frame containing all labels of a wekafile.
#'@examples
#'labels <- eccCLwekaLabels(weka)
eccCLwekaLabels <- function(wekaLoader) {
  numInstances <- .jcall(wekaLoader, "I", "getNumInstances")
  numLabels <- .jcall(wekaLoader, "I", "getNumLabels")
  labels <- .jcall(wekaLoader, "[D", "getLabels")
  dim(labels) <- c(numInstances, numLabels)
  labelFrame = as.data.frame(labels)
  return (labels)
}

#'@title Returns the number of instances contained in a wekafile.
#'@description Returns the number of instances contained in a wekafile.
#'@param wekaLoader A wekafile loaded e.g. with \code{\link{eccCLloadWekaFile}}.
#'@return The number of instances contained in a wekafile.
#'@examples
#'numInstances <- eccClwekaNumInstances(weka)
eccCLwekaNumInstances <- function(wekaLoader) {
  numInstances <- .jcall(wekaLoader, "I", "getNumInstances")
  return (numInstances)
}

#'@title Builds an ECC-classifier
#'@description Builds an ECC-classifier
#'@param values The attributes to use as data.frame.
#'@param labels The labels to use as data.frame.
#'@param numInstances Number of instances to build with as integer.
#'@param ensembleSize (Optional) The size of the ECC as integer. In other words: the number of CCs to build.
#'@param forestSize (Optional) The number of Decision Trees to use for each Random Forest  as integer.
#'@param ensembleSubSetFraction (Optional) The size of a random subset as fraction (double)
#'of the original set of data to use for each CC.
#'@param forestSubSetFraction (Optional) The size of a random subset as a fraction (double)
#'of the ensemble set of data to use for each RF.
#'@param maxLevel (Optional) The maximum level/depth for a decision tree as integer.
#'@param maxAttributes (Optional) The number of attributes to consider each node while building a tree as integer.
#'@return The ECC-classifier as a Java-object.
#'@examples
#'ecc <- eccCLbuild(values, labels, numInstances, ensembleSize = as.integer(64))
eccCLbuild <- function(
  values, #data frame
  labels, #data frame
  numInstances, #integer,
  ensembleSize = as.integer(-1),
  forestSize = as.integer(-1),
  ensembleSubSetFraction = as.double(-1),
  forestSubSetFraction = as.double(-1),
  maxLevel = as.integer(-1),
  maxAttributes = as.integer(-1)
) {

  vals = as.numeric(unlist(values))
  labs = as.numeric(unlist(labels))

  valArray = .jarray(t(as.vector(vals)))
  labArray = .jarray(t(as.vector(labs)))

  ecc <- .jnew("ecccl.RBridge")

  .jcall(ecc, "V", "setData", valArray, labArray, numInstances)
  .jcall(ecc, "V", "setClassifierSizes", ensembleSize, forestSize, ensembleSubSetFraction, forestSubSetFraction, maxLevel, maxAttributes)
  .jcall(ecc, "V", "buildClassifier")

  return(ecc)

}

#'@title Builds an ECC-classifier from a wekafile.
#'@description Builds an ECC-classifier from a wekafile.
#'The parameters values, labels and numInstances are automatically extracted from the object.
#'@param weka A wekafile loaded e.g. with \code{\link{eccCLloadWekaFile}}.
#'@param ensembleSize (Optional) The size of the ECC as integer. In other words: the number of CCs to build.
#'@param forestSize (Optional) The number of Decision Trees to use for each Random Forest  as integer.
#'@param ensembleSubSetFraction (Optional) The size of a random subset as fraction (double)
#'of the original set of data to use for each CC.
#'@param forestSubSetFraction (Optional) The size of a random subset as a fraction (double)
#'of the ensemble set of data to use for each RF.
#'@param maxLevel (Optional) The maximum level/depth for a decision tree as integer.
#'@param maxAttributes (Optional) The number of attributes to consider each node while building a tree as integer.
#'@return The ECC-classifier as a Java-object.
#'@examples
#'ecc <- eccCLbuildFromObject(weka)
eccCLbuildFromObject <- function(
  weka,
  ensembleSize = as.integer(-1),
  forestSize = as.integer(-1),
  ensembleSubSetFraction = as.double(-1),
  forestSubSetFraction = as.double(-1),
  maxLevel = as.integer(-1),
  maxAttributes = as.integer(-1)
)
{
  vals = eccCLwekaValues(weka)
  labs = eccCLwekaLabels(weka)
  inst = eccCLwekaNumInstances(weka)

  ecc <- eccCLbuild(vals, labs, inst, ensembleSize, forestSize, ensembleSubSetFraction, forestSubSetFraction, maxLevel, maxAttributes)

  return (ecc)
}

#'@title Builds an ECC-classifier from a weka-formatted file.
#'@description Loads a wekafile, extracts values, labels and number of instances out of it and builds an ECC-classifier.
#'@param file Filename and path like in \code{\link{eccCLloadWekaFile}}.
#'@param ensembleSize (Optional) The size of the ECC as integer. In other words: the number of CCs to build.
#'@param forestSize (Optional) The number of Decision Trees to use for each Random Forest  as integer.
#'@param ensembleSubSetFraction (Optional) The size of a random subset as fraction (double)
#'of the original set of data to use for each CC.
#'@param forestSubSetFraction (Optional) The size of a random subset as a fraction (double)
#'of the ensemble set of data to use for each RF.
#'@param maxLevel (Optional) The maximum level/depth for a decision tree as integer.
#'@param maxAttributes (Optional) The number of attributes to consider each node while building a tree as integer.
#'@return The ECC-classifier as a Java-object.
#'@examples
#'ecc <- eccCLbuildFromFile("/home/user/FILE")
eccCLbuildFromFile <- function(
  file,
  ensembleSize = as.integer(-1),
  forestSize = as.integer(-1),
  ensembleSubSetFraction = as.double(-1),
  forestSubSetFraction = as.double(-1),
  maxLevel = as.integer(-1),
  maxAttributes = as.integer(-1)
)
{
  weka <- eccCLloadWekaFile(file)

  ecc <- eccCLbuildFromObject(weka, ensembleSize, forestSize, ensembleSubSetFraction, forestSubSetFraction, maxLevel, maxAttributes)

  return (ecc)
}

#'@title Loads a previously saved ECC-classifier.
#'@description Loads a previously saved ECC-classifier.
#'@param file The filename and path to the classifier to load as String.
#'@return The ECC-classifier.
#'@examples
#'ecc <- eccCLload("/home/user/ecc")
eccCLload <- function(file) {
  ecc <- .jnew("ecccl.RBridge")

  .jcall(ecc, "V", "loadClassifier", file)

  return (ecc)
}

#'@title Saves a ECC-classifier into a file.
#'@description Saves a ECC-classifier into a file.
#'@param ecc The classifier to save.
#'@param file Filename and path to save the classifier into.
#'@examples
#'eccCLstore(ecc, "/home/user/ecc")
eccCLstore <- function(ecc, file) {
  .jcall(ecc, "V", "saveClassifier", file)
}

#'@title Classifies some data using a given classifer.
#'@description Classifies some data using a given classifer.
#'@param ecc The classifier to use.
#'@param values A data.frame containing the data to classify.
#'@param labels A data.frame containing the labels for the data to classify.
#'@param numInstances The number of instances to classify as integer.
#'@param threshold (Optional) The threshold as double to determine, when a specific result is considered positive or negative.
#'@param instPerRun (Optional) Number of instances to classify in one run.
#'Maybe useful if the total amount of instances is too much for the GPU to handle in one run.
#'@return A Java-object containing the results.
#'@examples
#'results <- eccCLclassify(ecc, values, labels, numInstances)
eccCLclassify <- function(
  ecc, #ecc classifier (java object)
  values, #data frame
  labels, #data frame
  numInstances, #integer,
  threshold = 0.5,
  instPerRun = numInstances
) {

  valArray = .jarray(t(as.vector(values)))
  labArray = .jarray(t(as.vector(labels)))

  .jcall(ecc, "V", "setData", valArray, labArray, numInstances)

  container <- .jnew("ecccl.ECCOutputContainer")
  .jcall(container, "V", "setThreshold", threshold)
  .jcall(ecc, "V", "classify", container, as.integer(instPerRun))
  return (container)

}

#'@title Classifies some data from a wekafile using a given classifer.
#'@description Classifies some data using a given classifer.
#'@param ecc The classifier to use.
#'@param weka The wekafile containing the data to classify, e.g. loaded with \code{\link{eccCLloadWekaFile}}.
#'@param threshold (Optional) The threshold as double to determine, when a specific result is considered positive or negative.
#'@param instPerRun (Optional) Number of instances to classify in one run.
#'Maybe useful if the total amount of instances is too much for the GPU to handle in one run.
#'@return A Java-object containing the results.
#'@examples
#'results <- eccCLclassify(ecc, weka)
eccCLclassifyObject <- function(
  ecc, #ecc classifier (java object)
  weka,
  threshold = 0.5,
  instPerRun = -1
) {
  vals <- eccCLwekaValues(weka)
  labs <- eccCLwekaLabels(weka)
  inst <- eccCLwekaNumInstances(weka)

  if(instPerRun < 0) {
    instPerRun = inst;
  }

  results <- eccCLclassify(ecc, vals, labs, inst, threshold, instPerRun)

  return (results)
}

#'@title Classifies some data from a file using a given classifer.
#'@description Classifies some data using a given classifer.
#'@param ecc The classifier to use.
#'@param file The filename and path to the file containing the data to classify, like in \code{\link{eccCLloadWekaFile}}.
#'@param threshold (Optional) The threshold as double to determine, when a specific result is considered positive or negative.
#'@param instPerRun (Optional) Number of instances to classify in one run.
#'Maybe useful if the total amount of instances is too much for the GPU to handle in one run.
#'@return A Java-object containing the results.
#'@examples
#'results <- eccCLclassify(ecc, "/home/user/FILE")
eccCLclassifyFile <- function(
  ecc, #ecc classifier (java object)
  file,
  threshold = 0.5,
  instPerRun = -1
)
{
  weka <- eccCLloadWekaFile(file)

  results <- eccCLclassifyObject(ecc, weka, threshold, instPerRun)

  return (results)
}

#'@title Converts the result-container (Java-object) to a data.frame.
#'@description Converts the result-container (Java-object) to a data.frame.
#'The data.frame is one-dimensional and contains the results for each label of the first instance, results for each label of the second
#'instance ...
#'@param output The Java-object containing the results from classification.
#'@return The results as a data.frame (double values).
#'@examples
#'final <- eccCLgetResults(results)
eccCLgetResults <- function(output)
{
  results <- .jcall(output, "[D", "getResults")

  return (results)
}

#'@title Converts the result-container (Java-object) to a data.frame.
#'@description Converts the result-container (Java-object) to a data.frame.
#'The data.frame is one-dimensional and contains the results for each label of the first instance, results for each label of the second
#'instance ...
#'@param output The Java-object containing the results from classification.
#'@return The results as a data.fram (boolean values).
#'@examples
#'final <- eccCLgetResults(results)
eccCLgetBipartition <- function(output) {
  results <- .jcall(output, "[Z", "getBipartition")

  return (results)
}

#'@title Performs a cross validation.
#'@description Performs a cross validation and returns the results for each repetition, k, label and measure.
#'@param values The attributes to use as a data.frame.
#'@param labels The labels to use as a data.frame.
#'@param numInstances The number of instances to use as integer.
#'@param repetitions The number of repitions to perform as integer.
#'@param buildSize The size of the data set to use for building a classifier.
#'@param testSize The size of the data set to use for testing the classifier.
#'@param k The k-parameter for k-fold-crossvalidation as integer.
#'@param ensembleSize (Optional) The size of the ECC as integer. In other words: the number of CCs to build.
#'@param forestSize (Optional) The number of Decision Trees to use for each Random Forest  as integer.
#'@param ensembleSubSetFraction (Optional) The size of a random subset as fraction (double)
#'of the original set of data to use for each CC.
#'@param forestSubSetFraction (Optional) The size of a random subset as a fraction (double)
#'of the ensemble set of data to use for each RF.
#'@param maxLevel (Optional) The maximum level/depth for a decision tree as integer.
#'@param maxAttributes (Optional) The number of attributes to consider each node while building a tree as integer.
#'@return The results for each label and measure.
#'@examples
#'crResults <- eccCLcrossValidate(values, labels, numInstances, 3, 667, 333, 3)
eccCLcrossValidate <- function(
  values, #data frame
  labels, #data frame
  numInstances, #integer
  repetitions, #integer
  buildSize, #integer
  testSize, #integer
  k, #integer
  ensembleSize = as.integer(-1),
  forestSize = as.integer(-1),
  ensembleSubSetFraction = as.double(-1),
  forestSubSetFraction = as.double(-1),
  maxLevel = as.integer(-1),
  maxAttributes = as.integer(-1)
) {

  evalMode = as.integer(2)

  vals = as.numeric(unlist(values))
  labs = as.numeric(unlist(labels))

  valArray = .jarray(t(as.vector(vals)))
  labArray = .jarray(t(as.vector(labs)))

  ecc <- .jnew("ecccl.RBridge")

  .jcall(ecc, "V", "setData", valArray, labArray, numInstances)
  .jcall(ecc, "V", "setEvaluationParameters", evalMode, repetitions)
  .jcall(ecc, "V", "setDataSizes", buildSize, testSize, k)
  .jcall(ecc, "V", "setClassifierSizes", ensembleSize, forestSize, ensembleSubSetFraction, forestSubSetFraction, maxLevel, maxAttributes)
  .jcall(ecc, "V", "buildEvaluator")

  .jcall(ecc, "V", "run")

  results <- .jcall(ecc, "[D", "resultsByRepetitionK")

  colLength <- k * repetitions
  rowLength <- .jcall(ecc, "I", "resultRowLength")
  colNames <- .jcall(ecc, "[S", "resultColNames")
  rowNames <- .jcall(ecc, "[S", "resultRowNames")

  dim(results) <- c(colLength, rowLength)

  resultFrame = as.data.frame(results)
  names(resultFrame) <- colNames
  row.names(resultFrame) <- rowNames

  return (resultFrame)

}
